/* This file is part of the KDE project
   Copyright (C) 2004 Martin Ellis <martin.ellis@kdemail.net>
   Copyright (C) 2004 Jarosław Staniek <staniek@kde.org>

   This file has been automatically generated from
   tools/sql_keywords.sh and src/parser/KDbSqlScanner.l
   and tools/kdb_keywords.txt.

   Please edit the sql_keywords.sh, not this file!

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
*/

#include "KDbDriver_p.h"

const char* const KDbDriverPrivate::kdbSQLKeywords[] = {
    "AND",
    "AS",
    "CREATE",
    "FROM",
    "IN",
    "INTEGER",
    "IS",
    "JOIN",
    "LEFT",
    "LIKE",
    "NOT",
    "NULL",
    "ON",
    "OR",
    "RIGHT",
    "SELECT",
    "SIMILAR",
    "TABLE",
    "TO",
    "WHERE",
    "XOR",
    "AFTER",
    "ALL",
    "ASC",
    "BEFORE",
    "BEGIN",
    "BETWEEN",
    "BY",
    "CASCADE",
    "CASE",
    "CHECK",
    "COLLATE",
    "COMMIT",
    "CONSTRAINT",
    "CROSS",
    "DATABASE",
    "DEFAULT",
    "DELETE",
    "DESC",
    "DISTINCT",
    "DROP",
    "END",
    "ELSE",
    "EXPLAIN",
    "FOR",
    "FOREIGN",
    "FULL",
    "GROUP",
    "HAVING",
    "IGNORE",
    "INDEX",
    "INNER",
    "INSERT",
    "INTO",
    "KEY",
    "LIMIT",
    "MATCH",
    "NATURAL",
    "OFFSET",
    "ORDER",
    "OUTER",
    "PRIMARY",
    "REFERENCES",
    "REPLACE",
    "RESTRICT",
    "ROLLBACK",
    "ROW",
    "SET",
    "TEMPORARY",
    "THEN",
    "TRANSACTION",
    "UNION",
    "UNIQUE",
    "UPDATE",
    "USING",
    "VALUES",
    "WHEN",
    nullptr
};
