/* Transliteration table of 65535 unicode characters
   Do not edit this file, it is generated
   by generate_transliteration_table.sh script. */

#define TRANSLITERATION_TABLE_SIZE 65535
extern const char *const transliteration_table[TRANSLITERATION_TABLE_SIZE + 1];
