/* This file is part of the KDE project
   Copyright (C) 2004 Martin Ellis <martin.ellis@kdemail.net>
   Copyright (C) 2004 Jarosław Staniek <staniek@kde.org>
   Copyright (C) 2008 Sharan Rao <sharanrao@gmail.com>

   Copied from sqlite/sqlitekeywords.h

   This file has been automatically generated from
   tools/sql_keywords/sql_keywords.sh and
   ../../3rdparty/kexisql3/src/tokenize.c.

   Please edit the sql_keywords.sh, not this file!

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
*/

#include <xbaseDriver.h>

  const char* xBaseDriver::keywords[] = {
    "ABORT",
    "ATTACH",
    "CLUSTER",
    "CONFLICT",
    "DEFERRED",
    "DEFERRABLE",
    "DETACH",
    "EACH",
    "EXCEPT",
    "FAIL",
    "GLOB",
    "IMMEDIATE",
    "INITIALLY",
    "INSTEAD",
    "INTERSECT",
    "ISNULL",
    "NOTNULL",
    "OF",
    "PRAGMA",
    "RAISE",
    "STATEMENT",
    "TEMP",
    "TRIGGER",
    "VACUUM",
    "VIEW",
    0
  };
