/* This file is part of the KDE project
   Copyright (C) 2004 Martin Ellis <martin.ellis@kdemail.net>
   Copyright (C) 2004 Jarosław Staniek <staniek@kde.org>

   This file has been automatically generated from
   tools/sql_keywords/sql_keywords.sh and
   postgresql-7.4.6/src/backend/parser/keywords.c.

   Please edit the sql_keywords.sh, not this file!

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
*/

#include "PostgresqlDriver.h"

const char* PostgresqlDriver::m_keywords[] = {
    "ABORT",
    "ABSOLUTE",
    "ACCESS",
    "ACTION",
    "ADD",
    "AGGREGATE",
    "ALTER",
    "ANALYSE",
    "ANALYZE",
    "ANY",
    "ARRAY",
    "ASSERTION",
    "ASSIGNMENT",
    "AT",
    "AUTHORIZATION",
    "BACKWARD",
    "BIGINT",
    "BINARY",
    "BIT",
    "BOOLEAN",
    "BOTH",
    "CACHE",
    "CALLED",
    "CAST",
    "CHAIN",
    "CHAR",
    "CHARACTER",
    "CHARACTERISTICS",
    "CHECKPOINT",
    "CLASS",
    "CLOSE",
    "CLUSTER",
    "COALESCE",
    "COLUMN",
    "COMMENT",
    "COMMITTED",
    "CONSTRAINTS",
    "CONVERSION",
    "CONVERT",
    "COPY",
    "CREATEDB",
    "CREATEUSER",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "CURSOR",
    "CYCLE",
    "DAY",
    "DEALLOCATE",
    "DEC",
    "DECIMAL",
    "DECLARE",
    "DEFAULTS",
    "DEFERRABLE",
    "DEFERRED",
    "DEFINER",
    "DELIMITER",
    "DELIMITERS",
    "DO",
    "DOMAIN",
    "DOUBLE",
    "EACH",
    "ENCODING",
    "ENCRYPTED",
    "ESCAPE",
    "EXCEPT",
    "EXCLUDING",
    "EXCLUSIVE",
    "EXECUTE",
    "EXISTS",
    "EXTERNAL",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FIRST",
    "FLOAT",
    "FORCE",
    "FORWARD",
    "FREEZE",
    "FUNCTION",
    "GLOBAL",
    "GRANT",
    "HANDLER",
    "HOLD",
    "HOUR",
    "ILIKE",
    "IMMEDIATE",
    "IMMUTABLE",
    "IMPLICIT",
    "INCLUDING",
    "INCREMENT",
    "INHERITS",
    "INITIALLY",
    "INOUT",
    "INPUT",
    "INSENSITIVE",
    "INSTEAD",
    "INT",
    "INTERSECT",
    "INTERVAL",
    "INVOKER",
    "ISNULL",
    "ISOLATION",
    "LANCOMPILER",
    "LANGUAGE",
    "LAST",
    "LEADING",
    "LEVEL",
    "LISTEN",
    "LOAD",
    "LOCAL",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "LOCATION",
    "LOCK",
    "MAXVALUE",
    "MINUTE",
    "MINVALUE",
    "MODE",
    "MONTH",
    "MOVE",
    "NAMES",
    "NATIONAL",
    "NCHAR",
    "NEW",
    "NEXT",
    "NO",
    "NOCREATEDB",
    "NOCREATEUSER",
    "NONE",
    "NOTHING",
    "NOTIFY",
    "NOTNULL",
    "NULLIF",
    "NUMERIC",
    "OF",
    "OFF",
    "OIDS",
    "OLD",
    "ONLY",
    "OPERATOR",
    "OPTION",
    "OUT",
    "OVERLAPS",
    "OVERLAY",
    "OWNER",
    "PARTIAL",
    "PASSWORD",
    "PATH",
    "PENDANT",
    "PLACING",
    "POSITION",
    "PRECISION",
    "PREPARE",
    "PRESERVE",
    "PRIOR",
    "PRIVILEGES",
    "PROCEDURAL",
    "PROCEDURE",
    "READ",
    "REAL",
    "RECHECK",
    "REINDEX",
    "RELATIVE",
    "RENAME",
    "RESET",
    "RESTART",
    "RETURNS",
    "REVOKE",
    "ROWS",
    "RULE",
    "SCHEMA",
    "SCROLL",
    "SECOND",
    "SECURITY",
    "SEQUENCE",
    "SERIALIZABLE",
    "SESSION",
    "SESSION_USER",
    "SETOF",
    "SHARE",
    "SHOW",
    "SIMPLE",
    "SMALLINT",
    "SOME",
    "STABLE",
    "START",
    "STATEMENT",
    "STATISTICS",
    "STDIN",
    "STDOUT",
    "STORAGE",
    "STRICT",
    "SUBSTRING",
    "SYSID",
    "TEMP",
    "TEMPLATE",
    "TIME",
    "TIMESTAMP",
    "TOAST",
    "TRAILING",
    "TREAT",
    "TRIGGER",
    "TRIM",
    "TRUE",
    "TRUNCATE",
    "TRUSTED",
    "TYPE",
    "UNENCRYPTED",
    "UNKNOWN",
    "UNLISTEN",
    "UNTIL",
    "USAGE",
    "USER",
    "VACUUM",
    "VALID",
    "VALIDATOR",
    "VARCHAR",
    "VARYING",
    "VERBOSE",
    "VERSION",
    "VIEW",
    "VOLATILE",
    "WITH",
    "WITHOUT",
    "WORK",
    "WRITE",
    "YEAR",
    "ZONE",
    nullptr
};
